import threading
from time import sleep

import torch

from wth.utils import StopWatch
from wth.utils import cache_wrapper
from wth.torch.utils import get_device_local, set_device_local, try_gpu, set_dtype_local, get_dtype_local


def set_device(device):
    set_device_local(device)


thread1 = threading.Thread(target=set_device, args=[try_gpu()])
thread2 = threading.Thread(target=set_device, args=[try_gpu()])

thread1.start()
thread2.start()


def set_dtype(dtype):
    set_dtype_local(dtype)
    print(get_dtype_local())


thread1 = threading.Thread(target=set_dtype, args=[torch.float32])
thread2 = threading.Thread(target=set_dtype, args=[torch.float])

thread1.start()
thread2.start()



import torch

from .utils import try_gpu, set_device_local, set_dtype_local

import json
import sys

import yaml


class Config:
    """
    环境配置，使用 yml 文件。
    例如:
    ```
    active: local
      school-server:
      local:
        epochs: 10
        batch-size: 64
        dtype: float32
        cuda_no: 1
        dataset_path: /your/dataset/path
    ```
    active 必需标识启用的环境。
    后续使用 key: 环境配置的方式可以定义多个环境。
    """

    def __init__(self, config_file='config.yml'):
        """
        构造函数。
        Args:
            config_file: 配置文件所在位置，默认 config.yml。
        """
        with open(config_file) as f:
            self.original = yaml.load(f, Loader=yaml.FullLoader)
            self.active_profile = self.original['active']
            if self.active_profile is None:
                raise ValueError("No active profile")

            print(f"Active profile[{self.active_profile}].")
            sys.stdout.flush()

            self.args = self.original[self.active_profile]  # type: dict
            if self.args is None:
                print(f"Active profile[{self.active_profile}] has no arguments.", file=sys.stderr)
                sys.stderr.flush()
                self.args = {}
            self.epochs = self.args.get('epochs', 10)
            self.batch_size = self.args.get('batch-size', 64)
            self.dtype = self.args.get('dtype', 'float32')
            self.cuda_no = self.args.get('cuda-no')
            self.dataset_path = self.args.get('dataset-path')

        # 设置设备
        if self.cuda_no is not None:
            set_device_local(try_gpu(self.cuda_no))
        # 设置 dtype
        if self.dtype is not None:
            if self.dtype == 'float32':
                set_dtype_local(torch.float32)
            else:
                set_dtype_local(torch.float64)

    def __getitem__(self, item):
        if self.args.__contains__(item):
            return self.args[item]
        key = item.replace("_", "-")
        if self.args.__contains__(key):
            return self.args[key]
        return None

    def __str__(self):
        return json.dumps(self.original)



from typing import Optional
import threading

import torch

from torch import nn

# 线程保存的 device
_device_local = threading.local()


def try_gpu(device_no: Optional[int] = None) -> torch.device:
    """
    尝试获取 GPU 设备。
    先获取 cuda，没有选择 mps，还没有使用 cpu。
    :param device_no: 设备号，只对 cuda 有效
    :return: GPU 设备
    """
    if torch.cuda.is_available():
        if device_no is not None and device_no < torch.cuda.device_count():
            #
            return torch.device(f'cuda:{device_no}')

        return torch.device('cuda')
    return torch.device('mps' if torch.mps else 'cpu')


def set_device_local(device: torch.device):
    """
    设置线程变量：训练用设备
    :param device: 训练用设备
    """
    _device_local.device = device
    print(f'设置线程变量 device = {device.type}:{device.index} [{threading.currentThread().name}]')


def get_device_local() -> torch.device:
    """
    获取线程变量：训练用设备
    :return: 训练用设备
    """
    device = getattr(_device_local, 'device', None)
    if device is None:
        device = try_gpu()
        set_device_local(device)
    return device


def set_dtype_local(dtype: torch.dtype):
    """
    设置线程变量：训练数据类型
    Args:
        dtype: 训练用数据类型
    """
    _device_local.dtype = dtype

    print(f'设置线程变量 dtype = {dtype} [{threading.currentThread().name}]')


def get_dtype_local():
    dtype = getattr(_device_local, 'dtype', None)
    if dtype is None:
        dtype = torch.float
        set_dtype_local(dtype)
        print(f'线程变量 dtype 为空，已设置为 {dtype}.')
    return dtype


def module_to(module: nn.Module) -> nn.Module:
    """
    将模型转换为线程的设备类型
    Args:
        module: 要转换的模型

    Returns:
        模型本身
    """
    return module.to(get_device_local())


def tensor_to(*data: torch.Tensor) -> list[torch.Tensor]:
    """
    将数据 tensor 或者 tensor 列表 转换为线程保存的设备类型。
    Args:
        data: 要转换的数据。

    Returns:
        数据的转换结果。
    """
    return [i.to(get_device_local()) for i in data]



import time

from prettytable import PrettyTable


class StopWatch:
    def __init__(self):
        self._times = []
        self._job_names = []
        self._start_time = None
        self._cur_job_name = None
        self._pt = PrettyTable(["Job", "Time", "Percent"])
        # self.pt.set_style(FRAME)

    def start(self, name=''):
        self._start_time = time.time()
        self._cur_job_name = name

    def stop(self):
        if self._cur_job_name is None:
            raise Exception("Job name is required")
        self._times.append(time.time() - self._start_time)
        self._job_names.append(self._cur_job_name)
        self._cur_job_name = None
        return self._times[-1]

    def avg(self):
        return self.sum() / len(self._times)

    def sum(self):
        return sum(self._times)

    def display(self):
        total = self.sum()
        for i in range(0, len(self._times)):
            self._pt.add_row([self._job_names[i], f'{self._times[i]:.3f}', f'{self._times[i] / total * 100:.2f}%'])
        self._pt.title = f"Sum: {total:.3f}, Avg: {self.avg():.3f}"
        print(self._pt)



from enum import Enum
from functools import wraps
import os.path


class CacheType(Enum):
    PICKLE = "pkl"
    JOBLIB = "lib"


def cache_wrapper(prefix, save_path=None, force_write=False, use: CacheType = CacheType.PICKLE):
    """
    装饰器，训练完保存模型，如果模型存在则加载。
    :param prefix: 保存的文件名
    :param force_write: 强行刷新模型
    :param save_path: 模型保存位置
    :param use: 使用的方法: pickle/joblib
    """

    # 外层为了获取 save_path 参数
    # wrapper 才是真正的装饰器
    def wrapper(train_func):
        """

        :param train_func: 训练模型的函数
        """

        @wraps(train_func)
        def inner(*args, **kwargs):
            args_append = "_".join([str(arg) for arg in args])
            kwargs_append = "_".join([f'{k}_{v}' for k, v in kwargs.items()])
            suffix = 'pkl' if use == CacheType.PICKLE else 'lib' if use == CacheType.JOBLIB else ''

            format_filename = f'{prefix}_{args_append}_{kwargs_append}.{suffix}'

            if save_path is not None:
                _try_make_dirs(save_path)
                real_filename = os.path.join(save_path, format_filename)
            else:
                real_filename = format_filename
            if force_write or not os.path.exists(real_filename):
                # 强制写入或者没有缓存
                # 执行并保存
                model = train_func(*args, **kwargs)
                if use == CacheType.JOBLIB:
                    import joblib
                    joblib.dump(model, real_filename)
                    print(f'任务完成, 保存{type(model)} -> {real_filename} ({use})')
                elif use == CacheType.PICKLE:
                    import pickle
                    with open(real_filename, 'wb') as f:
                        pickle.dump(model, f)
                        print(f'任务完成, 保存{type(model)} -> {real_filename} ({use})')
                else:
                    raise ValueError(f'未知的保存模型方式{use}')
                return model
            else:
                # 加载
                if use == CacheType.JOBLIB:
                    import joblib
                    ret = joblib.load(real_filename)
                    print(f'加载模型 {type(ret)} <-- "{real_filename}" ({use}).')
                    return ret
                if use == CacheType.PICKLE:
                    import pickle
                    with open(real_filename, 'rb') as f:
                        ret = pickle.load(f)
                        print(f'加载模型 {type(ret)} <-- "{real_filename}" ({use}).')
                        return ret

                raise ValueError(f'未知的保存模型方式{use}')

        return inner

    return wrapper


def _try_make_dirs(path):
    if not os.path.exists(path):
        # 创建文件夹
        os.makedirs(path)
        print(f"创建缓存文件夹: {path}")
